# -*- coding: UTF-8 -*-
'''
Feature definitions for ANTARES
'''

from __future__ import unicode_literals
__author__     = 'Gautham Narayan'
__email__      = 'gsnarayan@gmail.com'
__maintainer__ = 'Gautham Narayan'
__license__    = 'GPLv3'
__version__    = '0.1'
__status__     = 'alpha'
